import socket
import time

from lib import status

class DownloadError(Exception):
    """download()の呼び出し元にダウンロード処理失敗を知らせる例外"""
    pass

class AgainError(Exception):
    """対象ホストのモードを切り替えたので、直後にもう一度実行する必要がある場合に発生する例外"""
    pass

class FailError(Exception):
    """実行に失敗した時に発生する例外"""
    pass

class TCPDownloadProtocol:
    
    def __init__(self, imageFile, targetHost):
        self._imageFile = imageFile
        self._targetHost = targetHost
        self._socket = None
    
    def download(self):
        self._targetHost.hostStatus = status.HostStatus.IDLE
        
        againCountMax = 3
        for i in range(againCountMax):
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    self._socket = s
                    s.connect(self._targetHost.getHostPort())
                    
                    # initial command
                    dataToSend = self._sendTestMode()
                    
                    dataRecv = bytearray()
                    
                    # send data and wait for response from targetHost
                    while dataToSend is not None:
                        self._socket.sendall(dataToSend)
                        
                        while len(dataRecv) < 8:
                            # TODO 一定時間来なかった場合の処理
                            dataRecv2 = s.recv(8)
                            # 既存のデータの最後に受信したデータを追加する
                            dataRecv += dataRecv2
                        
                        response = dataRecv[0:8]
                        dataRecv = dataRecv[8:]
                        dataToSend = self._processResponse(response)
                    
                    s.shutdown(socket.SHUT_RDWR)
                    break
            except AgainError:
                continue
            except socket.timeout:
                raise DownloadError("接続に失敗しました")
    
    def _processResponse(self, response):
        """対象ホストからの返答を処理する"""
        # Windows版(ver.3)におけるCTcpdwlDlg::ProcessPendingRead()に相当
        ackHead = b'\x52\x03'
        nakHead = b'\x52\x06'
        modeHead = b'\x52\x05'
        
        if response.startswith(ackHead):
            return self._processACK(response)
        elif response.startswith(nakHead):
            return self._processNAK(response)
        elif response.startswith(modeHead):
            return self._processModeNotify(response)
        else:
            # 受信データがおかしい
            head = response[0:2]
            raise FailError("対象ホストからのコマンドコード異常: {0}".format(head.hex()))

    def _processACK(self, response):
        """対象ホストからACKが帰ってきた時"""
        state = self._targetHost.hostStatus
        
        if state == status.HostStatus.SWITCH:
            # 最初からやり直す
            raise AgainError()
        elif state == status.HostStatus.STARTDOWN:
            self._targetHost.hostStatus = status.HostStatus.DOWNLOAD
            return self._sendChunk(0) # 最初のchunkを送る
        elif state == status.HostStatus.DOWNLOAD:
            self._targetHost.confirmSentChunk()
            if self._targetHost.chunkSent == self._imageFile.getMaxChunkIndex():
                # 最後のchunkに対するACKが帰ってきた
                self._targetHost.hostStatus = status.HostStatus.FINISHED
                return None # 処理終了
            else:
                # まだ途中である
                return self._sendChunk(self._targetHost.getNextChunk()) # 次のchunkを送る
        else:
            # 状態がおかしい
            raise FailError("ACK処理エラー 内部状態: {0}".format(state))

    def _processNAK(self, response):
        """対象ホストからNAKが帰ってきた時"""
        errCode = response[2] << 8 | response[3]
        
        if errCode & 0x8000 != 0:
            errStr = "NAK(Command error)"
        elif errCode & 0x0040 != 0:
            errStr = "無効なファームウェアです"
        elif errCode & 0x0020 != 0:
            errStr = "NAK(Download size error)"
        elif errCode & 0x0010 != 0:
            errStr = "NAK(Dial-in/Dial-out is used)"
        elif errCode & 0x0008 != 0:
            errStr = "NAK(Telnet is used)"
        elif errCode & 0x0004 != 0:
            errStr = "NAK(Flash ROM erase error)"
        elif errCode & 0x0002 != 0:
            errStr = "NAK(Download in progress)"
        elif errCode & 0x0001 != 0:
            errStr = "NAK(Flash ROM write error)"
        else:
            errStr = "NAK(unknown code 0x{0:x})".format(errCode)
        
        raise FailError(errStr)

    def _processModeNotify(self, response):
        """対象ホストからモード通知が帰ってきた時"""
        rInq1 = b'\x52\x05\x01\x00\x00\x00\x00\x00'
        rInq2 = b'\x52\x05\x00\x01\x00\x00\x00\x00'
        rInq3 = b'\x52\x05\x02\x00\x00\x00\x00\x00'
        rInq4 = b'\x52\x05\x00\x02\x00\x00\x00\x00'
        rInq5 = b'\x52\x05\x03\x00\x00\x00\x00\x00'
        
        if response == rInq1 or response == rInq2:
            self._targetHost.hostStatus = status.HostStatus.SWITCH
            return self._sendSwitchMode()
        elif response == rInq3 or response == rInq4 or response == rInq5:
            self._targetHost.hostStatus = status.HostStatus.STARTDOWN
            return self._sendStartDownload()
        else:
            # 何かおかしい
            raise FailError("MODE処理エラー")

    def _sendTestMode(self):
        """テストモードコマンド文字列を送信する"""
        return b'\x51\x21\x00\x00\x00\x00\x00\x00'

    def _sendSwitchMode(self):
        """対象ホストにSwitchコマンドを送信する"""
        return b'\x51\x20SWITCH'

    def _sendStartDownload(self):
        fileSize = self._imageFile.getSize()
        command = bytearray(8)
        command[0] = 0x51
        command[1] = 0x22
        command[2] = (fileSize & 0xFF000000) >> 24
        command[3] = (fileSize & 0x00FF0000) >> 16
        command[4] = (fileSize & 0x0000FF00) >> 8
        command[5] = (fileSize & 0x000000FF)
        # 6, 7は0x00のまま
        
        return command

    def _sendChunk(self, index):
        """指定されたindexのchunkを対象ホストに送信する"""
        data = self._imageFile.getChunk(index)
        self._targetHost.chunkSent = index
        return data