#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# TCP Downloader for Linux
# (c) 2017 Century Systems Co., Ltd.

import argparse
from logging import basicConfig, getLogger, DEBUG, INFO, WARNING, ERROR, CRITICAL
import os
import sys

from lib import exe
from lib import targethost
from lib import imgfile

version='1.0'

def loginit(logLevelStr):
    global logger
    logLevelDict = {
        'DEBUG': DEBUG,
        'INFO': INFO,
        'WARNING': WARNING,
        'ERROR': ERROR,
        'CRITICAL': CRITICAL,
    }
    logLevel=logLevelDict.get(logLevelStr, INFO)
    basicConfig(level=logLevel)
    logger = getLogger(__name__)

def parse_args():
    global version
    parser = argparse.ArgumentParser(
        description='TCP downloader for Century Systems products. (version {0})'.format(version))
    # 引数やオプションを解釈する部分
    parser.add_argument('--loglevel', metavar='LEVEL',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help='log level (DEBUG|INFO|WARNING|ERROR|CRITICAL) (default: INFO)',
        default='INFO')
    parser.add_argument('-c', '--concurrent', type=int, metavar='NUM',
        help='Concurrent Execution Limit (default=20, max=100)', default=20)
    parser.add_argument('-e', '--errorfile', metavar='ERRORFILE',
        help='output of error host:port list (default: errorfile.txt)',
        default='errorfile.txt')
    parser.add_argument('-l', '--listfile', action='store_true',
        help='ARG is file name of IP address list (default: ARG is <IP address>[:<port>])')
    parser.add_argument('ARG',
        help='<IP address>[:<port>] or filename')
    parser.add_argument('IMG',
        help='firmware image file (*.img)')
    
    args = parser.parse_args()
    
    # concurrentがレンジ内かチェック
    # choicesを使うとエラーメッセージが冗長になるので使わない
    if args.concurrent < 1 or args.concurrent > 100:
        print("concurrent(-c / --concurrent) must in 1-100.")
        sys.exit(os.EX_USAGE)
    
    loginit(args.loglevel)
    # 解釈した引数を返す
    return args

def getListFromFile(filename):
    try:
        return targethost.TargetHost.getTargetHosts(filename)
    except Exception as ex:
        print("error in reading list file.")
        sys.exit(os.EX_USAGE)
    
def readImageFile(filename):
    try:
        imageFile = imgfile.ImageFile(filename)
    except Exception as ex:
        print("error in reading image file.")
        sys.exit(os.EX_DATAERR)

    return imageFile

def makeTargetHosts(args):
    concurrent = args.concurrent
    if args.listfile == True:
        (targetHosts, errorLines) = getListFromFile(args.ARG)
        if len(errorLines) > 0:
            for errorLine in errorLines:
                print("wrong target: {0}".format(errorLine))
            sys.exit(os.EX_USAGE)
    else:
        try:
            targetHost = targethost.TargetHost(args.ARG)
            targetHosts = [targetHost]
        except Exception:
            print("wrong target: {0}".format(args.ARG))
            sys.exit(os.EX_USAGE)
    
    return targetHosts

def openErrorFile(filename):
    try:
        errorFile = open(filename, 'x')
        return errorFile
    except FileExistsError:
        print("error file exist.")
        sys.exit(os.EX_CANTCREAT)

def doTCPDownload(args, imageFile, targetHosts, errorFile):
    global logger
    executor = exe.TCPDownload(args, logger=logger)
    return executor.execute(imageFile, targetHosts, errorFile)

def writeErrorFile(errorFileName, errorFile, errorList):
    if len(errorList) > 0:
        for errorHost in errorList:
            errorStr = errorHost.getHostPortStr() + "\n"
            errorFile.write(errorStr)
        errorFile.close()
    else:
        errorFile.close()
        os.remove(errorFileName)

def main():
    args = parse_args()
    imageFile = readImageFile(args.IMG)
    targetHosts = makeTargetHosts(args)
    errorFile = openErrorFile(args.errorfile)
    errorList = doTCPDownload(args, imageFile, targetHosts, errorFile)
    writeErrorFile(args.errorfile, errorFile, errorList)

if __name__ == "__main__":
    main()
